# mypy: ignore-errors

import torch
import numpy as np
import pytest
from numpy.typing import NDArray

from typing import List, Tuple

from llguidance.numpy import get_bitmask_shape
import llguidance.numpy
import llguidance.torch

try:
    import llguidance.mlx as ll_mlx
except ImportError:
    ll_mlx = None # type: ignore

has_cuda = torch.cuda.is_available()
dev = "cuda" if has_cuda else "cpu"


def u32_to_bool(arr: NDArray[np.uint32]) -> NDArray[np.uint8]:
    arr = arr.astype(np.uint32)
    bool_arr = np.unpackbits(arr.view(np.uint8), bitorder="little").reshape(
        arr.shape[0], arr.shape[1] * 32
    )
    return bool_arr


def measure_gpu_time(func, *args, **kwargs): # type: ignore
    start = torch.cuda.Event(enable_timing=True)
    end = torch.cuda.Event(enable_timing=True)

    torch.cuda.synchronize()  # Ensure previous ops are finished
    start.record()

    func(*args, **kwargs)  # Run the function

    end.record()
    torch.cuda.synchronize()  # Wait for all ops to finish

    return start.elapsed_time(end)  # Returns time in milliseconds


def gen_test_data(
    batch: int, vocab: int, add_vocab: int
) -> Tuple[np.ndarray, np.ndarray]:
    data = np.random.rand(batch, vocab + add_vocab).astype(np.float32)
    mask = np.random.randint(0, 2**32, get_bitmask_shape(batch, vocab), dtype=np.uint32)
    mask = mask.astype(np.int32)
    return data, mask


def validate_mask_data(data_np: np.ndarray, mask_np: np.ndarray, md_np: np.ndarray):
    cutoff = min(32 * mask_np.shape[1], data_np.shape[1])
    bool_mask = u32_to_bool(mask_np)[:, :cutoff]
    md_cut = md_np[:, :cutoff]
    data_cut = data_np[:, :cutoff]
    assert np.all(md_cut[bool_mask == 0] == -np.inf)
    assert np.all(md_cut[bool_mask == 1] == data_cut[bool_mask == 1])
    assert np.all(md_np[:, cutoff:] == -np.inf)


def gen_test_datas():
    yield gen_test_data(1, 32000, 0)
    yield gen_test_data(3, 32000, 0)
    yield gen_test_data(10, 32000, 0)
    yield gen_test_data(3, 128000, 0)
    for v in range(32001, 32033):
        yield gen_test_data(3, v, 0)
        yield gen_test_data(3, v, 137)
    for v in range(1, 33):
        yield gen_test_data(3, 32007, v)


def run_test(cb):
    for data_np, mask_np in gen_test_datas():
        res = cb(data_np, mask_np)
        validate_mask_data(data_np, mask_np, res)


def test_mask_data_torch():
    def torch_test(data_np: np.ndarray, mask_np: np.ndarray):
        data = torch.tensor(data_np, device=dev)
        mask = torch.tensor(mask_np, device=dev)

        masked_data = data.clone()
        llguidance.torch.apply_token_bitmask_inplace(masked_data, mask)

        return masked_data.cpu().numpy()

    run_test(torch_test)


def test_mask_data_numpy():
    def numpy_test(data_np: np.ndarray, mask_np: np.ndarray):
        masked_data = data_np.copy()
        llguidance.numpy.apply_token_bitmask_inplace(masked_data, mask_np)
        return masked_data

    run_test(numpy_test)


def test_mask_data_mlx():
    if not ll_mlx:
        pytest.skip("mlx is not available")

    import mlx.core as mx

    def mlx_test(data_np: np.ndarray, mask_np: np.ndarray):
        assert ll_mlx
        data = mx.array(data_np)
        masked_data = ll_mlx.apply_token_bitmask(data, mask_np)
        return np.array(masked_data)

    run_test(mlx_test)


def test_mask_data_perf():
    if not has_cuda:
        pytest.skip("CUDA is not available")
    times = []
    for batch in [100, 105, 95, 107, 108, 99]:
        data_np, mask_np = gen_test_data(batch, 128000, 0)
        data = torch.tensor(data_np, device=dev)
        mask = torch.tensor(mask_np, device=dev)
        for _ in range(10):
            masked_data = data.clone()
            t = measure_gpu_time(
                llguidance.torch.apply_token_bitmask_inplace, masked_data, mask
            )
            times.append(round(t * 1000))
    print(times)


if __name__ == "__main__":
    test_mask_data_perf()



from typing import List, Optional, Sequence, Any
from ._util import TokenId


class TokenizerWrapper:
    eos_token_id: TokenId
    bos_token_id: Optional[TokenId]
    tokens: Sequence[bytes]
    special_token_ids: Sequence[int]

    def __init__(self, gtokenizer: Any) -> None:
        # these are required by LLTokenizer:
        self.eos_token_id = gtokenizer.eos_token_id
        self.bos_token_id = gtokenizer.bos_token_id
        self.tokens = gtokenizer.tokens
        self.special_token_ids = getattr(gtokenizer, "special_token_ids", [])
        self.is_tokenizer_wrapper = True

        # more private stuff
        self._gtokenizer = gtokenizer
        self._accepts_bytes = True
        try:
            gtokenizer(b"test")
        except:
            self._accepts_bytes = False

    def _encode_string(self, s: str) -> List[TokenId]:
        r: List[TokenId]
        if self._accepts_bytes:
            r = self._gtokenizer(s.encode("utf-8"))
        else:
            r = self._gtokenizer(s)
        return r

    # required by LLTokenizer
    __call__ = _encode_string



from typing import List, Union, Dict, Any
from dataclasses import dataclass
import json
import sys


class StructTag:
    """
    Represents a structural tag used for constraining function/tool invocations in the middle
    of running text.

    These tags are just syntactic sugar over lazy lexemes or special tokens.
    See also https://github.com/guidance-ai/llguidance/blob/main/docs/syntax.md#tool-calling

    Attributes:
        trigger (str): A substring or token that signals when to start applying this tag.
            For example "<function" or "<|python_tag|>"
        begin (str): The beginning of the tag. Must start with trigger.
            For example "<function=foo>" or '<|python_tag|>{"name":"foo","parameters":'
        grammar (Union[str, dict]): The grammar definition for the arguments of function.
            It can be JSON schema (stringified or as a dict),
            or a full Lark grammar (with 'start:' rule).
        end (str): The string to force at the end of the structured segment.
            For example: "</function>", "}", ""
    """

    def __init__(self, *, trigger: str, begin: str,
                 grammar: Union[str, Dict[str, Any]], end: str):
        self.trigger = trigger
        self.begin = begin
        self.grammar = grammar
        self.end = end
        self._validate()

    def _validate(self) -> None:
        assert len(self.trigger) > 0, "trigger must not be empty"
        assert self.begin.startswith(
            self.trigger), "begin must start with trigger"

    @staticmethod
    def to_grammar(
        tags: List['StructTag'],
        *,
        assume_special: bool = True,
        text_regex: str = r"(.|\n)*",
    ) -> str:
        """
        Generates a Lark grammar string based on the provided structural tags.

        Arguments:
        tags: List[StructTag]: A list of structural tags to generate grammar for.
        assume_special: bool: A flag indicating whether to assume triggers of the format <...> are in fact special tokens.
            Defaults to true.
        text_regex: str: A regex pattern for matching text segments, defaults to r"(.|\n)*", which allows all strings.
        """

        def gtext(s: str) -> str:
            if s:
                return json.dumps(s)
            else:
                return ""

        assert len(tags) > 0, "tags must not be empty"
        assert "/" not in text_regex, "text_regex must not contain /"
        for tag in tags:
            assert isinstance(tag,
                              StructTag), "tags must be StructTag instances"
            tag._validate()
        tag_options = " | ".join(f"tag_{i}" for i in range(len(tags)))
        lark = f"""
%llguidance {{}}
start: ({tag_options})* tag_end
tag_end: TAG_TEXT
TAG_TEXT: /{text_regex}/
"""
        side_grammars = []

        for tag_idx, tag in enumerate(tags):
            lark += "\n"
            tag_rule = f"tag_{tag_idx}"
            trig = tag.trigger

            if isinstance(tag.grammar, str):
                if tag.grammar.lstrip().startswith("{"):
                    grm = "%json " + tag.grammar
                else:
                    gname = f"{tag_rule}_grm"
                    side_grammars.append({
                        "name": gname,
                        "lark_grammar": tag.grammar
                    })
                    grm = "@" + gname
            elif isinstance(tag.grammar, dict):
                grm = "%json " + json.dumps(tag.grammar)
            else:
                raise ValueError("grammar must be a string or a dictionary")

            beg = tag.begin[len(trig):]
            body = f"{gtext(beg)} {grm} {gtext(tag.end)}"

            if assume_special and trig.startswith("<") and trig.endswith(">"):
                # f_qux: TEXT <|placeholder1|> "qux(" /[0-9]+/ ")"
                lark += f"{tag_rule}: TAG_TEXT {trig} {body}\n"
            else:
                # f_baz_hd[lazy]: TEXT "<tool"
                # f_baz: f_baz_hd "=baz>" /[0-9]+/ "</tool>"
                lark += f"{tag_rule}_trig[lazy]: TAG_TEXT {gtext(trig)}\n"
                lark += f"{tag_rule}: {tag_rule}_trig {body}\n"

        lark = lark.lstrip()
        if side_grammars:
            side_grammars.insert(0, {
                "name": "struct_tag",
                "lark_grammar": lark
            })
            return json.dumps({"grammars": side_grammars})
        else:
            return lark
